/*
	CTaskMil - Task execute parameters, including dialog for task progress 
	TaskMil( CVwTask * ) - 

*/

//-----------------------------------------------------------------------------
// options which are set by dialog and used by task.
class COptionsMilStimulus : public CItem { 
	public : 
		CString m_cName;
		int m_iChannel;			// if channel = -1, use m_dValue
		double m_dValue;
		bool m_bValue;
		COptionsMilStimulus( const char * pch, int iChannel, double dValue, bool bValue );
	};

const int OPT_MODE_EVALUATE = 0;
const int OPT_MODE_OPTIMIZE = 1;

class COptionsMil { 
	public : 
		//-- User options
		int m_iDataSelect;		//0=whole file; 1=list; 2=current selection.
		int m_iDataChannel;
		CString m_cDataName;
		double  m_dDeadTime;
		int m_iChannels;
	
		CItemArray m_aStim;
		COptionsMilStimulus * Stimulus( int iStim );

		int m_iMode;			// evaluate / Optimize / ... 

		//-- Optimization options
		int m_iMaxIter;			// maximum dfp iterations
		double m_dConvLL;		// LL convergence criteria
		double m_dConvGrad;		// Gradient convergence criteria
		bool m_bNumericDerivatives;	// Use analytical or numeric derivatives
		double m_dK0Step;
		double m_dK1Step;
		
		//-- inputs
		CModel m_oModel;
		CModel * m_pOriginalModel;
		CEventList m_oEventList;
		
		double m_dSampling;
		
		//-- Results
		int m_iIterations;		// actual iterations
		double m_dLL;			// final log likelihood

		//-- 
		COptionsMil();
		~COptionsMil();		
		//int BufferSize();
		//void ToBuffer( BYTE * buf );
		//void FromBuffer( BYTE * buf );
		#ifdef WIN32
		bool RegistryLoad( );
		bool RegistrySave( );
		#endif
	};

//-----------------------------------------------------------------------------
using namespace TNT;

class CMilModelData {					// collection of the stimulus dependent MIL data 
	public : 
		Array1D<double> m_X;
		Array1D<double> m_Z;			// from m_X
		Array2D<double> m_Mtx;			// from m_Z
		Array1D<double> m_Vct;			// from m_Z
		Array2D<double> m_dx_z;			// from m_Mtx

		Array2D<double> m_dqq_q;		// from MetaModel.QQ in ModelCalculate
		Array2D<double> m_aQQe;
		Array3D<double> m_dQQe_qq;
		Array2D<double> m_dq_x;
		Array2D<double> m_aW;			// FreePar() of MetaModel.QQ in ModelCalculate
		Array3D<double> m_adP;			// FreePar() of MetaModel.QQ in ModelCalculate
		Array3D<double> m_adP1;			// FreePar() of MetaModel.QQ in ModelCalculate

		//-- 
		Array2D<double> m_df_qqe;
		Array1D<double> m_df_qq;
		Array1D<double> m_df_q;
		Array1D<double> m_df_x;
		Array1D<double> m_df_z;

		//double m_dMagnitude;
	};
		
class CTaskMil : public CVwTask { 
	public : 
		COptionsMil * m_pOptionsMil;
		HWND m_hModelWnd;

		CTaskMil();
		virtual ~CTaskMil();
		bool m_bAnalyticalDerivative;

	public : 
		//-- (Meta) Model
		void ModelSetup( CModel * pModel, CMilModelData * pData );
		void ModelCalculate( CModel * pModel, double dDeadTime, CMilModelData * pData );
		CMilModelData * m_pData;		// array of 1 ModelData for each stimulus set 
		//double m_dMagnitudeSum;
		
		void ModelMultipleSetup( COptionsMil * pOpt );	// Model setup / calculate for all stimulii sets

		CMetaModel m_oMetaModel;
		int m_iMetaPaths;
		int m_iMetaClasses;
		int m_iMetaStates;
		int m_iOrigRates;
		int m_iOrigPaths;
		int m_iXSize;
		int m_iZSize;
		Array2D<int> m_aiPaths;
		Array1D<int> m_aiClasses;		// from metamodel.UsedClassCount
		Array1D<int> m_aiStateClasses;
		Array1D<double> m_adInitPr;

		//-- Workspace
		void EvaluateSetup( int iEvents );
		Array2D<double> m_adAlpha;
		Array2D<double> m_adBeta;
		Array1D<double> m_adScale;

		//-- Calculations
		void mil_backward( int iDwells, short * aiDwell, int * aiDuration, double dSampling );
		void mil_expQ( int iDim, double * w, double ** p, double ** p1,  double t, double** a);
		double mil_gama(  double dEig1, double dEig2, double dTDwell );
		void mil_dlogLL_qe( );
		void qtoqe( int iStates, int iPaths, Array2D<int> & path, Array1D<int> & aiClass, 
				Array1D<int> & aiStateClass, int iClasses, double td, CMilModelData * pData, 
				Array2D<double> & q );
		void qabc(	int na, int nb, int nc, 
			Array1D<int> & ida, Array1D<int> & idb, Array1D<int> & idc, 
			Array2D<double> & qab, Array2D<double> & qbb, Array2D<double> & qbc, 
			double td, 
			Array2D<int> & path, int iPaths, Array2D<double> & qac, Array3D<double> * pdqac_q
			);
		bool ZtoX( CMilModelData * pData, double * z, bool bUpdateRates ) ;

		double EvaluateLL( );
		void EvaluateGradient( double * adGradient );
		void OptimizeAnalytical( );
		
		double EvaluateStep( int iRate, int iDir, int iK, double dStep );
		double EvaluateSteps( double * aStep, int iRates, double dInitLL );
		void OptimizeNumerical( );

	};

//-----------------------------------------------------------------
DWORD WINAPI TaskMil( LPVOID pParam );

